"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.savePage = exports.fetchPage = exports.fixBaseURI = exports.fetchMedia = exports.loadPage = exports.getPage = void 0;
const tslib_1 = require("tslib");
const lodash_memoize_1 = tslib_1.__importDefault(require("lodash.memoize"));
const puppeteer_core_1 = tslib_1.__importDefault(require("puppeteer-core"));
const jsdom_1 = require("jsdom");
const node_fetch_1 = tslib_1.__importDefault(require("node-fetch"));
const path_1 = require("path");
const node_toolkit_1 = require("@tech_query/node-toolkit");
const yaml_1 = require("yaml");
const fs_extra_1 = require("fs-extra");
const config_1 = require("./config");
const parser_1 = require("./parser");
exports.getPage = (0, lodash_memoize_1.default)(async () => {
    const browser = await puppeteer_core_1.default.launch({
        executablePath: config_1.executablePath,
        args: ['--no-sandbox']
    });
    const page = await browser.newPage();
    await Promise.all([
        page.setRequestInterception(true),
        page.setUserAgent(config_1.userAgent)
    ]);
    page.setDefaultNavigationTimeout(0);
    page.on('request', request => {
        if (['image', 'media'].includes(request.resourceType()))
            request.abort();
        else
            request.continue();
    });
    return page;
});
async function loadPage(URI, root_selector = 'body') {
    if (!config_1.executablePath)
        return jsdom_1.JSDOM.fromURL(URI, {
            userAgent: config_1.userAgent,
            pretendToBeVisual: true,
            runScripts: 'outside-only'
        });
    const page = await (0, exports.getPage)();
    await page.goto(URI);
    await page.waitForSelector(`${root_selector}:not(:empty)`);
    return new jsdom_1.JSDOM(await page.content(), { url: URI });
}
exports.loadPage = loadPage;
const MediaSelector = 'img, audio, video', LinkSelector = 'a[href], area[href]';
async function fetchMedia(root, root_path = '') {
    const list = Array.from(root.querySelectorAll(MediaSelector), async (media) => {
        try {
            var { pathname, protocol, href } = (0, parser_1.sourcePathOf)(media);
        }
        catch (_a) {
            media.remove();
        }
        if (protocol === 'data:')
            var { data } = await (0, node_toolkit_1.blobFrom)(href);
        else
            var data = await (await (0, node_fetch_1.default)(href, { headers: { 'User-Agent': config_1.userAgent } })).buffer();
        const name = await (0, parser_1.createFilePath)(data, pathname, root_path);
        media.src = name;
        if (media.style.display === 'none')
            media.style.display = '';
        return { name, data };
    });
    return (await Promise.allSettled(list))
        .map(result => result.status === 'fulfilled'
        ? result.value
        : console.error(result.reason))
        .filter(Boolean);
}
exports.fetchMedia = fetchMedia;
function fixBaseURI(document, baseURI) {
    const base = document.createElement('base');
    base.href = baseURI;
    document.head.append(base);
    for (const media of document.querySelectorAll(MediaSelector))
        media.setAttribute('src', media.src);
    for (const link of document.querySelectorAll(LinkSelector))
        link.setAttribute('href', link.href);
    base.remove();
}
exports.fixBaseURI = fixBaseURI;
async function fetchPage({ source, rootSelector, baseURI = '' }) {
    const { window: { document } } = await loadPage(source);
    let root;
    for (const selector of [rootSelector, ...config_1.body_tag])
        if (selector && (root = document.querySelector(selector)))
            break;
    const meta = (0, parser_1.parsePage)(document, config_1.meta_tag), media = await fetchMedia(root, (0, path_1.parse)(source).name);
    if (baseURI)
        fixBaseURI(document, baseURI);
    for (const element of document.querySelectorAll(parser_1.IgnoredTags + ''))
        element.remove();
    for (const element of document.querySelectorAll('[style*="display:"]')) {
        const { display, visibility, opacity, width, height } = element.style;
        if (display === 'none' ||
            visibility === 'hidden' ||
            !(opacity || width || height))
            element.remove();
    }
    return Object.assign(Object.assign({}, meta), { content: root.innerHTML.trim().replace(/(\n\s*)+/g, '\n'), media });
}
exports.fetchPage = fetchPage;
async function savePage({ source, rootSelector, baseURI = '', markdown, rootFolder = process.cwd() }) {
    console.time('Fetch');
    const _a = await fetchPage({
        source,
        rootSelector,
        baseURI
    }), { content, media } = _a, meta = tslib_1.__rest(_a, ["content", "media"]);
    const root_path = (0, path_1.join)(rootFolder, ...(meta.categories || []).map(parser_1.fileNameOf));
    const text_file = (0, path_1.join)(root_path, `${(0, parser_1.fileNameOf)((0, path_1.parse)(source).name)}.${markdown ? 'md' : 'html'}`), data = !markdown
        ? content
        : `---
${(0, yaml_1.stringify)(meta)}
---

${parser_1.convertor.turndown(content)}`;
    await (0, fs_extra_1.outputFile)(text_file, data);
    console.log('[save] ' + text_file);
    for (const { name, data } of media) {
        const media_file = (0, path_1.join)(root_path, name);
        await (0, fs_extra_1.outputFile)(media_file, data);
        console.log('[save] ' + media_file);
    }
    console.log('--------------------');
    console.timeEnd('Fetch');
}
exports.savePage = savePage;
